# Value Function Iteration to solve the stochastic  
# Single-State Optimal Growth model 
#
# Original AMPL coding: June 30, 2006

# important parameters
param beta := 0.8;     # discount factor
param n := 11;         # number of Chebyshev nodes
param deg := 10;       # degree of Chebyshev polynomial
param kmin := 0.5;     # minimum state value
param kmax := 1.5;     # maximum state value

# Define some constants
param pi  := 3.14159265358979;
param eps := 1e-6;

# run time
param runTime default 0;

# Define parameters values
set   SN := 1..n;       # index set of states
set   Sdeg := 1..(deg+1);
param alpha := 0.25;
param A := (1-beta)/(alpha*beta); 
param maxL := 3;
param B := (maxL-1)*(1-alpha);

# Define Chebyshev nodes
param z{i in SN} := - cos((2*i-1)/(2*n)*pi); # the Chebyshev interpolation nodes on [-1,1]
param ext := -(z[1]+1)*(kmax-kmin)/2/z[1];
param extkmin := kmin - ext;
param extkmax := kmax + ext;
param dz := 2/(extkmax-extkmin); # dz/dx 
param KNodes {i in SN} := extkmin + (1+z[i])/dz;  # set of Chebyshev state variables

# Define variables
var coefs {i in Sdeg}; # coefficients
var kPlus {i in SN}; # tomorrow capital
var L {i in SN} >= eps; # labor supply
var c {i in SN} >= eps; # consumptions

# model for Bellman
maximize TV : sum {i in SN} ( (log(c[i]) + B*log(maxL-L[i])) + beta*(sum {j in Sdeg} 
              coefs[j]*cos((j-1)*acos(dz*(kPlus[i]-extkmin)-1))) );
subject to  Balance {i in SN} : kPlus[i] = KNodes[i] + A*(KNodes[i]^alpha)*(L[i]^(1-alpha)) - c[i];
subject to  kL {i in SN} : kPlus[i] >= kmin;
subject to  kU {i in SN} : kPlus[i] <= kmax;
subject to  LU {i in SN} : L[i] <= maxL-eps;

# define Chebyshev basis polynomial and their derivatives values at KNodes 
# by use of the recursive formula
param T {i in SN, j in Sdeg} default 0; # Chebyshev polynomial values at KNodes
param dT {i in SN, j in Sdeg} default 0; # Chebyshev polynomial first derivative values at KNodes
param ddT {i in SN, j in Sdeg} default 0; # Chebyshev polynomial second derivative values at KNodes
for {i in SN} {
  let T[i,1] := 1;
  let T[i,2] := z[i]; 
  let dT[i,1] := 0;
  let dT[i,2] := 1 * dz; 
  let ddT[i,1] := 0;
  let ddT[i,2] := 0;
  for {j in 2..deg} {
    let T[i,j+1] := 2*z[i]*T[i,j] - T[i,j-1];
    let dT[i,j+1] := 2*T[i,j]*dz + 2*z[i]*dT[i,j] - dT[i,j-1];
    let ddT[i,j+1] := 4*dT[i,j]*dz + 2*z[i]*ddT[i,j] - ddT[i,j-1];
  }
}


param V {i in SN} default 0; # current value function's values at KNodes

# model for Shaping 
minimize LeastSQ_a: sum {i in SN} ( (sum {j in Sdeg} (coefs[j]*T[i,j])) - V[i] )^2;
subject to Vgrad {i in SN} : sum {j in 2..(deg+1)} coefs[j]*dT[i,j] >= eps;
subject to Vhess {i in SN} : sum {j in 3..(deg+1)} coefs[j]*ddT[i,j] <= -eps;

######################################

problem Bellman: TV, kPlus, L, c, Balance, kL, kU, LU;
problem Shaping: coefs, LeastSQ_a, Vgrad, Vhess;

######################################

param MAX_count default 500;
param counter, default 0;
param err default 1.0;
param err2 default 1.0;

# Define Uniform state values
param  un   :=  107;        # number of states
set    SUN   := 1..un;       # index set of states
param  UKNodes {i in SUN} := kmin + (kmax-kmin)/(un-1)*(i-1);  # set of states 

param Va_plus {k in SUN}; # tomorrow value function's values at UKNodes
param Va {k in SUN}; # current value function's values at UKNodes

# define Chebyshev basis polynomial values at UKNodes 
# by use of the recursive formula
param Uz {i in SUN} := dz*(UKNodes[i]-extkmin) - 1;  # normalize UKNodes
param UT {i in SUN, j in Sdeg} default 0; # Chebyshev polynomial values at UKNodes
for {i in SUN} {
  let UT[i,1] := 1;
  let UT[i,2] := Uz[i]; 
  for {j in 2..deg} {
    let UT[i,j+1] := 2*Uz[i]*UT[i,j] - UT[i,j-1];
  }
}

let {i in Sdeg} coefs[i] := 0;
let {k in SUN} Va[k] := sum {j in Sdeg} coefs[j]*UT[k,j];

repeat { 
  option minos_options 'Superbasics_limit=500'; 
  solve Bellman;
  let runTime := runTime + _solve_time;
  let {i in SN} V[i] := (log(c[i]) + B*log(maxL-L[i])) + beta*(sum {j in Sdeg} 
              coefs[j]*cos((j-1)*acos(dz*(kPlus[i]-extkmin)-1))); 

  option minos_options 'Superbasics_limit=500'; 
  solve Shaping;
  let runTime := runTime + _solve_time;

  let {k in SUN} Va_plus[k] := Va[k];
  let {k in SUN} Va[k] := sum {j in Sdeg} coefs[j]*UT[k,j];
  let err := max {k in SUN} abs(Va_plus[k]-Va[k]);
  let err2 := sqrt(sum {k in SUN} (Va_plus[k]-Va[k])^2);
  if (counter mod 10 = 0) then { 
    display counter, LeastSQ_a, err, err2;
  }
  let counter := counter + 1;
} while ( err > eps) && (counter < MAX_count);

let runTime := runTime + _ampl_time;

display Va;

display counter, runTime, LeastSQ_a, err, err2, coefs;

# compute slopes and curvatures at nodes
param slope {i in SN};
let {i in SN} slope[i] := sum {j in 2..(deg+1)} coefs[j]*dT[i,j];
param curvature {i in SN};
let {i in SN} curvature[i] := sum {j in 3..(deg+1)} coefs[j]*ddT[i,j];

display {i in SN} (KNodes[i], kPlus[i], c[i], L[i], V[i], slope[i], curvature[i]); 

########## END
